/******************************************************************************
 * lightmanager.h 
 * implements the management of all lights within the scene
 * Copyright (C) 2010 Marc Gissler <gisslerm@informatik.uni-freiburg.de>
 ******************************************************************************/

//TODO: currently, all lights are scene-aligned -> allow camera-aligned lights, too

#ifndef LIGHTMANAGER_H
#define LIGHTMANAGER_H

#include "globals.h"

#include "gmMath\gmVec3.h"
#include "gmMath\gmVec4.h"

#include <vector>

// ------------------------------------------------------------
// forward declarations
// 

class CShadow;

// ------------------------------------------------------------
// defines thes light class
// 

class CLight
{
public:	
	GLenum m_lightId;

	GLfloat m_pos[4]; // if w == 0, then its a direction!
	
	bool m_castShadows;

	GLfloat m_ambientColor[3];
	GLfloat m_diffuseColor[3];
	GLfloat m_specularColor[3];
	GLfloat m_shadowColor[3];

	virtual void draw() = 0;
	virtual void position() = 0;

	void setAmbientIntensity();
	void setAllIntensitiesButAmbient();
	void setAllIntensities();
};

// ------------------------------------------------------------
// derives a point light class
// 

class CPointLight : public CLight
{
public:
	CPointLight(GLenum lightId, gmVector4& pos, gmVector3& ambient, gmVector3& diffuse, gmVector3& specular);

	virtual void draw();
	virtual void position();	
};

// ------------------------------------------------------------
// derives a spotlight class
// 

class CSpotLight : public CLight
{
public:
	GLfloat		m_dir[3];
	GLfloat		m_cutOff[1];

	CSpotLight(GLenum lightId, gmVector4& pos, gmVector3& direction, float cutOff, gmVector3& ambient, gmVector3& diffuse, gmVector3& specular);

	virtual void draw();
	virtual void position();
};

// ------------------------------------------------------------
// defines thes light manager class
// 

class CLightManager
{
	//constructor/destructor
	CLightManager();
	~CLightManager();
	
	// member variables
	static CLightManager* m_pInstance;

	std::vector<CPointLight*> m_pointLights;
	std::vector<CSpotLight*> m_spotLights;
	int m_maxLights;

	CShadow* m_pShadow;

	void clear();

public:
	// destroys the single light manager
	static void destroyLightManager();

	// draws the scene-aligned lights in the form of spheres
	void drawSceneAlignedLights();
	
	// draws the scene with shadows
	void drawSceneWithShadows();

	// gets the single light manager
	static CLightManager* getLightManager();

	// adds a point or directional (pos.w == 0) light to the scene
	//	(returns the lightId)
	int addPointLight(gmVector4& pos, gmVector3& ambient, gmVector3& diffuse, gmVector3& specular);

	// adds a spot light to the scene
	//	(returns the lightId)
	int addSpotLight(gmVector4& pos, gmVector3& direction, float cutOff, gmVector3& ambient, gmVector3& diffuse, gmVector3& specular);

	// returns a specific light
	CLight* getLight(int lightId);

	// returns the lights in the scene
	std::vector<CLight*> getLights();

	// returns, whether a light is enabled/disabled
	bool isLightEnabled(int lightId);

	// positions the scene-aligned lights
	void positionSceneAlignedLights();

	// toggles the light on/off
	void toggleLight(int lightId, bool toggle);

	// sets the ambient term of the intensity exlusively
	void setAmbientIntensity();

	// sets all intensities but the ambient term
	void setAllIntensitiesButAmbient();

	// sets all intensities as given in the add[...]Light-methods
	void setAllIntensities();

	// sets the shadow-casting capabilities of a light
	void setHasShadow(int lightId, bool hasShadow);
};
#endif
